
use std::collections::HashSet;
use std::path::PathBuf;
use vulkano::pipeline::GraphicsPipelineAbstract;
use vulkano::framebuffer::RenderPassAbstract;
use std::sync::Arc;

/*

Realistically, what should the API for this thing look like...

It's going to just generate a pipeline. But that consists of loading and compiling various shaders,
    and generating a pipeline for those shaders and other customer behaviour.

This best works I think if I allow users to
    A.) impl from a base trait which allows resource lookup
    B.) Generate 1 of each of the types of shaders
    C.) Modify specilization constants, whatever that might mean
    D.) impl from a base trait which defines it's interface

*/



/// Inheriting this gives private functions to grab resources
pub(super) trait CompiledGraphicsPipelineResources {
    fn get_paths(filename: String, types: HashSet<ShaderType>) -> Vec<(ShaderType, PathBuf)> {
        let project_root =
            std::env::current_dir()
                .expect("failed to get root directory");

        let mut shader_path = project_root.clone();
        shader_path.push(PathBuf::from("resources/shaders/"));

        let mut paths = Vec::new();

        for shader_type in types {
            match shader_type {
                ShaderType::VERTEX => {
                    let mut shader_path = shader_path.clone();
                    shader_path.push(PathBuf::from(filename.clone() + ".vert"));
                    paths.push((shader_type, shader_path));
                }
                ShaderType::FRAGMENT => {
                    let mut shader_path = shader_path.clone();
                    shader_path.push(PathBuf::from(filename.clone() + ".frag"));
                    paths.push((shader_type, shader_path));
                }
                ShaderType::GEOMETRY => {
                    let mut shader_path = shader_path.clone();
                    shader_path.push(PathBuf::from(filename.clone() + ".geom"));
                    paths.push((shader_type, shader_path));
                }
                ShaderType::TESSELLATION_CONTROL => {
                    let mut shader_path = shader_path.clone();
                    shader_path.push(PathBuf::from(filename.clone() + ".tesscont"));
                    paths.push((shader_type, shader_path));
                }
                ShaderType::TESSELLATION_EVALUATION => {
                    let mut shader_path = shader_path.clone();
                    shader_path.push(PathBuf::from(filename.clone() + ".tesseval"));
                    paths.push((shader_type, shader_path));
                }
            }
        }

        paths
    }
}

/// Typed wrapper for a u32 handle
#[derive(Clone, Debug, Default, PartialEq, Eq, Hash)]
pub struct CompiledGraphicsPipelineHandle {
    pub handle: u32
}

pub trait CompiledGraphicsPipeline {
    fn get_name(&self) -> String;
    fn get_handle(&self) -> Arc<CompiledGraphicsPipelineHandle>;
    fn get_pipeline(&self) -> Arc<dyn GraphicsPipelineAbstract + Sync + Send>;
    fn recompile(self, render_pass: Arc<dyn RenderPassAbstract + Send + Sync>)
                 -> Self where Self: Sized;
}

/// Legacy ShaderType enum for single type shaders.
#[derive(PartialEq, Eq, Hash, Clone)]
pub enum ShaderType {
    VERTEX = 0,
    FRAGMENT = 1,
    GEOMETRY = 2,
    TESSELLATION_CONTROL = 3,
    TESSELLATION_EVALUATION = 4,
}